/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * export - gtk handlers & code for exporting a plot as an image
 */

#include "export.xh"
#include "util.xh"

static void export_image(my_plot_t *plot, const char *filename, const char *format, int width, int height);

typedef struct {
      GtkEntry *entry_reso_x;
      GtkEntry *entry_reso_y;
      GtkWidget *label_reso_times;

      GtkRadioButton *radiobutton_reso_plot;

      my_plot_t *plot;

} eac_callback_data_t;

static void eac_reso(eac_callback_data_t *data, const char *w, const char *h) {
   gtk_entry_set_text(data->entry_reso_x, w);
   gtk_entry_set_text(data->entry_reso_y, h);
   gtk_widget_set_sensitive(GTK_WIDGET(data->entry_reso_x), 0);
   gtk_widget_set_sensitive(GTK_WIDGET(data->entry_reso_y), 0);
   gtk_widget_set_sensitive(GTK_WIDGET(data->label_reso_times), 0);
}

static void eac_reso_640x480(GtkWidget *button, eac_callback_data_t *data) {
   eac_reso(data, "640", "480");
}

static void eac_reso_800x600(GtkWidget *button, eac_callback_data_t *data) {
   eac_reso(data, "800", "600");
}

static void eac_reso_1024x768(GtkWidget *button, eac_callback_data_t *data) {
   eac_reso(data, "1024", "768");
}

static void eac_reso_plot(GtkWidget *button, eac_callback_data_t *data) {
   char w[6], h[6];
   snprintf(w, sizeof(w), "%d", data->plot->da->allocation.width);
   snprintf(h, sizeof(h), "%d", data->plot->da->allocation.height);
   eac_reso(data, w, h);
}

static gboolean eac_plot_configure_event(GtkWidget *widget, GdkEventConfigure *do_not_use, eac_callback_data_t *data) {
   if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(data->radiobutton_reso_plot))) {
      eac_reso_plot(NULL, data);
   }
   return 0;
}

static void eac_reso_custom(GtkWidget *button, eac_callback_data_t *data) {
   gtk_widget_set_sensitive(GTK_WIDGET(data->entry_reso_x), 1);
   gtk_widget_set_sensitive(GTK_WIDGET(data->entry_reso_y), 1);
   gtk_widget_set_sensitive(GTK_WIDGET(data->label_reso_times), 1);
}

void export_as_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *data) {
   GladeXML *gladexml = glade_xml_new(data->gladefile, "filechooserdialog_export_as", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "filechooserdialog_export_as");

/*
   GtkWidget* dialog = gtk_file_chooser_dialog_new("Export image as",
						   data->parent,
						   GTK_FILE_CHOOSER_ACTION_SAVE,
						   "gtk-cancel", GTK_RESPONSE_CANCEL,
						   "gtk-save-as", GTK_RESPONSE_ACCEPT,
						   NULL);
*/
   eac_callback_data_t eaccbd = {
      .entry_reso_x = GTK_ENTRY(glade_xml_get_widget(gladexml, "entry_reso_x")),
      .entry_reso_y = GTK_ENTRY(glade_xml_get_widget(gladexml, "entry_reso_y")),
      .label_reso_times = glade_xml_get_widget(gladexml, "label_reso_times"),

      .radiobutton_reso_plot = GTK_RADIO_BUTTON(glade_xml_get_widget(gladexml, "radiobutton_reso_plot")),

      .plot = data->plot,
   };


   g_signal_connect(glade_xml_get_widget(gladexml, "radiobutton_reso_640x480"), "toggled", G_CALLBACK(eac_reso_640x480), &eaccbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "radiobutton_reso_800x600"), "toggled", G_CALLBACK(eac_reso_800x600), &eaccbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "radiobutton_reso_1024x768"), "toggled", G_CALLBACK(eac_reso_1024x768), &eaccbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "radiobutton_reso_plot"), "toggled", G_CALLBACK(eac_reso_plot), &eaccbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "radiobutton_reso_custom"), "toggled", G_CALLBACK(eac_reso_custom), &eaccbd);

   gulong ce_handler = g_signal_connect(data->plot->da, "configure-event", G_CALLBACK(eac_plot_configure_event), &eaccbd);

   hide_markers(data->plot);

   GtkFileChooser *chooser = GTK_FILE_CHOOSER(dialog);

   // 2.6 or 2.8 ?
#if GTK_CHECK_VERSION(2,6,0)
   GtkFileFilter *filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, "Image formats");
   gtk_file_filter_add_pixbuf_formats(filter);
   gtk_file_chooser_add_filter(chooser, filter);
#endif

   if(gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
      char *filename = gtk_file_chooser_get_filename(chooser);
      int w = atoi(gtk_entry_get_text(eaccbd.entry_reso_x));
      int h = atoi(gtk_entry_get_text(eaccbd.entry_reso_y));
      if(w < 1 || h < 1) {
	 fprintf(stderr, "Invalid size %dx%d\n", w, h);
      } else {
	 export_image(data->plot, filename, "png", w, h);
      }
      g_free (filename);
   }

   g_signal_handler_disconnect(data->plot->da, ce_handler);
   gtk_widget_destroy (dialog);
}

static void export_image(my_plot_t *plot, const char *filename, const char *format, int width, int height) {
   printf("Save plot to %s\n", filename);
   GdkPixmap *pixmap = gdk_pixmap_new(plot->da->window, width, height, -1);
   GdkGC *gc = gdk_gc_new_with_fg_color(pixmap, "black");
   x_gdk_draw_rectangle(pixmap, gc, 1, 0, 0, width, height);
   plot_draw_to(plot, pixmap, 0);
   GdkPixbuf *pixbuf = gdk_pixbuf_get_from_drawable(NULL, pixmap, NULL, 0, 0, 0, 0, -1, -1);
   g_object_unref(pixmap);
   GError *error = NULL;
   if(!gdk_pixbuf_save(pixbuf, filename, "png", &error, NULL)) {
      fprintf(stderr, "Unable to export image: %s\n", error ? error->message : NULL);
   }
   g_object_unref(pixbuf);
}
