/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot_draw - handles the drawing of a plot
 */

#include <math.h>

#include "plot_draw.xh"
#include "util.xh"
#include "grid.xh"

#if 0
#define pprintf printf
#else
#define pprintf(...)
#endif

static void plot_draw_legend(my_plot_t *plot, GdkDrawable *dest, gboolean show_disabled, PangoFontDescription *font_desc, int w, int h);

static gint8 major_dashes[] = { 2, 2 };
static gint8 minor_dashes[] = { 1, 3 };

// also see DA_ALLOCATION in plot2.xc
#define DA_ALLOCATION_DEST int w, h; gdk_drawable_get_size(dest, &w, &h); h -= 24;

void plot_draw_to(my_plot_t *plot, GdkDrawable *dest, gboolean show_disabled) {
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);

   DA_ALLOCATION_DEST;

   // setup gc
   // ##TODO##
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;
      for(int i=0; i<line_count; ++i) {
	 if(line[i].line_gc) continue;
	 // printf("%d %s\n", i, line[i].color);

	 line[i].line_gc = gdk_gc_new_with_fg_color(dest, line[i].color);
	 line[i].horiz_marker = gdk_gc_new_with_fg_color(dest, line[i].color);

	 gint8 horiz_marker_dashes[] = { 2, 4 };
	 gdk_gc_set_dashes(line[i].horiz_marker, 0, horiz_marker_dashes, SZ(horiz_marker_dashes));
	 gdk_gc_set_line_attributes(line[i].horiz_marker, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
	 gdk_gc_set_function(line[i].horiz_marker, GDK_XOR);
      }
   }

   if(!plot->minor_gc) {
      plot->minor_gc = gdk_gc_new_with_fg_color(dest, plot->minor_color);
      plot->major_gc = gdk_gc_new_with_fg_color(dest, plot->major_color);
      plot->sep_gc = gdk_gc_new_with_fg_color(dest, plot->sep_color);
      plot->label_gc = gdk_gc_new_with_fg_color(dest, plot->label_color);
      plot->cursor_marker_gc = gdk_gc_new_with_fg_color(dest, plot->cursor_marker_color);

      gdk_gc_set_line_attributes(plot->minor_gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
      gdk_gc_set_line_attributes(plot->major_gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
      gdk_gc_set_line_attributes(plot->cursor_marker_gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);

      gdk_gc_set_function(plot->cursor_marker_gc, GDK_XOR);

      //	foreach my $family ($da->get_pango_context()->get_font_map()->list_families()) {
      //	    print %{$family}."\n";
      //	}

      plot->legend_bg_gc = gdk_gc_new_with_fg_color(dest, plot->legend_bg_color);
      plot->legend_border_gc = gdk_gc_new_with_fg_color(dest, plot->legend_border_color);
      plot->legend_text_gc = gdk_gc_new_with_fg_color(dest, plot->legend_text_color);
   }

   hide_markers(plot);
   plot->box = COORD_UNDEF;

   // my $queued = 0;

   //    gtk_main_iteration() while(gtk_events_pending());
   //    if($mytrans != plot->dsc_x->{num_changes}) {
   //	print "- $mytrans plot->dsc_x->{num_changes}\n";
   //#	    $da->queue_draw() unless($queued++);
   //	$conc--; return;
   //    }

   //############# grids

   PangoFontDescription *font_desc = pango_font_description_from_string("Verdana 10");

   // value grid
   double hs1 = yoff;
   double he1 = yoff + ylength;

   pprintf("y %lf %lf\n", hs1, he1);

   grid_t hgrid = calc_value_grid(hs1, he1, h / 50, h);

   pprintf("%lf %lf %lf %lf %lf %lf %d\n", hgrid.dmin, hgrid.d, hgrid.dmax, hgrid.dvmin, hgrid.dv, hgrid.dvmax, hgrid.minpermaj);

   int doff;
   doff = (int)(xoff * w / xlength);
   gdk_gc_set_dashes(plot->major_gc, doff, major_dashes, SZ(major_dashes));
   gdk_gc_set_dashes(plot->minor_gc, doff, minor_dashes, SZ(minor_dashes));

   for(double yo = hgrid.dmin, yvo = hgrid.dvmin; yo < hgrid.dmax + hgrid.d; yo += hgrid.d, yvo += hgrid.dv) {
      if(yo < 0) continue;
      int y = h - (int)yo;
      char tmp[30];
      snprintf(tmp, sizeof(tmp), "%lf", yvo);
      PangoLayout *layout = gtk_widget_create_pango_layout(plot->da, tmp);
      pango_layout_set_font_description(layout, font_desc);
      gdk_draw_layout(dest, plot->label_gc, 0, y, layout);
      gdk_draw_line(dest, ((int)(round(yvo / hgrid.dv))) % hgrid.minpermaj == 0 ? plot->major_gc : plot->minor_gc, 0, y, w, y);
   }

   // time grid
   double ts1f = xoff;
   double te1f = xoff + xlength;

   pprintf("x %lf %lf\n", ts1f, te1f);

   grid_t vgrid = calc_time_grid(ts1f, te1f, w / 90, w);

   pprintf("%lf %lf %lf %lf %lf %lf %d\n", vgrid.dmin, vgrid.d, vgrid.dmax, vgrid.dvmin, vgrid.dv, vgrid.dvmax, vgrid.minpermaj);

   gdk_draw_line(dest, plot->sep_gc, 0, h, w, h);

   doff = -(int)(yoff * h / ylength);
   gdk_gc_set_dashes(plot->major_gc, doff, major_dashes, SZ(major_dashes));
   gdk_gc_set_dashes(plot->minor_gc, doff, minor_dashes, SZ(minor_dashes));

   int odd = ((int)round(vgrid.dvmin / vgrid.dv)) % 2;
   for(double xo = vgrid.dmin, xvo = vgrid.dvmin; xo < vgrid.dmax + vgrid.d; xo += vgrid.d, xvo += vgrid.dv) {
      odd = !odd;
      int x = (int)xo;
      PangoLayout *layout = gtk_widget_create_pango_layout(plot->da, format_time((my_time_t)round(xvo), 1));
      pango_layout_set_font_description(layout, font_desc);
      int lx, ly;
      pango_layout_get_pixel_size(layout, &lx, &ly);
      gdk_draw_layout(dest, plot->label_gc, x - lx/2, h + (odd ? 0 : 12), layout);
      gdk_draw_line(dest, ((int)round(xvo / vgrid.dv)) % vgrid.minpermaj == 0 ? plot->major_gc : plot->minor_gc, x, 0, x, h + (odd ? 0 : 10));
   }

   my_time_t ts1 = (my_time_t)ts1f;
   my_time_t te1 = (my_time_t)te1f;

   //############ data

   double w_div_xlength = w / xlength;
   double h_div_vdiff = h / ylength;

   GdkRectangle clip = { 0, 0, w, h };

#define MAX_POINTS 1000
#define POINTS_PER_X 2

   // allocate points array fulfilling the maximum need
   int maxlines = 0;
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      maxlines = max(maxlines, plot->line_set[lsi].line_count);
   }
   GdkPoint *points = malloc_c(maxlines * MAX_POINTS * sizeof(GdkPoint));
   int *usedPoints = malloc_c(maxlines * sizeof(int));
   double *y = malloc_c(maxlines * sizeof(double));
   gboolean *gotundef = malloc_c(maxlines * sizeof(gboolean));

   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->line_set[lsi].dataset;
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      int is = my_time_arr_find(dataset->time, ts1, 0);
      int ie = my_time_arr_find(dataset->time, te1, 1);

      //	if(gtk_events_pending()) {
      //	    $da->queue_draw() unless($queued++);
      //
      //	    $conc--; return;
      //	}

      for(int i=0; i<line_count; ++i) {
	 if(!line[i].enabled) continue;
	 //#TODO## rectangle ref
	 gdk_gc_set_clip_rectangle(line[i].line_gc, &clip);
	 line[i].data = dataset->item[line[i].dataset_idx].data;
	 pprintf("[%d] = %lf\n", i, dataset->item[i].max);
      }

      double x = (dataset->time[is] - xoff) * w_div_xlength;

      // how many data entries can be skipped per row. Starts from 1, approaches correct value during rendering
      int lc = 1;

      bzero(usedPoints, line_count * sizeof(int));
      bzero(gotundef, line_count * sizeof(gboolean));

      pprintf("ii %d %d\n", is, ie);

      for(int j=is ; j<=ie; ) {
	 int sx = (int)(x * POINTS_PER_X);
	 bzero(y, line_count * sizeof(double));
	 int js = j;
	 int c = 0;
	 while(1) {
	    for(int i=0; i<line_count; ++i) {
	       if(!line[i].enabled) continue;
	       double d = line[i].data[j];
	       if(d > 84) {
		  pprintf("t%d i%d %lf\n", j, i, d);
	       }
	       if(!__isnan(d)) {
		  switch(line[i].mode) {
		     case 0:
			y[i] += d;
			break;
		     case 1:
			if(c == 0)
			   y[i] = d;
			break;
		  }
	       } else {
		  gotundef[i] = 1;
	       }
	    }
	    j += lc;
	    c++;
	    if(j > ie) break;
	    x = (dataset->time[j] - xoff) * w_div_xlength; // ##TODO## this sometimes gives undef value in subtract
	    if((int)(x * POINTS_PER_X) > sx) break;
	 }

	 // adjust skip value
	 lc = (int)(c * lc / 4);
	 if(lc < 1) lc = 1;

	 if((dataset->time[js] - dataset->time[j]) / (js - j) > dataset->maxdiff) {
	    memset(gotundef, 1, line_count * sizeof(gboolean));
	 }

	 for(int i=0; i<line_count; ++i) {
	    if(!line[i].enabled) continue;
	    if(gotundef[i]) {
	       gotundef[i] = 0;
	       //pprintf("gun %d %d\n", i, usedPoints[i]);
	       if(usedPoints[i] > 1) {
		  pprintf("gu %d %d\n", i, usedPoints[i]);
		  gdk_draw_lines(dest, line[i].line_gc, &points[MAX_POINTS * i], usedPoints[i]);
	       }
	       usedPoints[i] = 0;
	    } else {
	       double yv;
	       switch(line[i].mode) {
		  case 0:
		     yv = c > 1 ? y[i] / c : y[i];
		     break;
		  case 1: {
		     int je = j < ie ? j : ie;
//		     yv = (line[i].data[je] - y[i]) * TIME_MULTIPLIER / (dataset->time[je] - dataset->time[js]);
		     yv = (line[i].data[je] - y[i]) * TIME_MULTIPLIER / (dataset->time[je] - dataset->time[js]) * (1 * TIME_MULTIPLIER);
		     break;
		     default:
			fprintf(stderr, "Bad mode %d in %d.%d\n", line[i].mode, lsi, i);
			return;
		  }
	       }

	       // calculate point coordinates
	       GdkPoint *point = &points[MAX_POINTS * i + usedPoints[i]++]; //##TODO##
	       point->x = sx / POINTS_PER_X;
	       int ly = (he1 - yv) * h_div_vdiff;

	       // with extreme overdrives, these tend to go wrong (or so I guess)
	       if(ly > 32767) ly = 32767;
	       else if(ly < -32768) ly = -32768;
	       point->y = ly;

	       if(usedPoints[i] == MAX_POINTS) {
		  pprintf("gm %d %d\n", i, usedPoints[i]);
		  gdk_draw_lines(dest, line[i].line_gc, &points[MAX_POINTS * i], usedPoints[i]);
		  points[MAX_POINTS * i] = points[MAX_POINTS * i + usedPoints[i]-1];
		  usedPoints[i] = 1;
	       }
	    }
	 }
      }

      // plot final lines
      for(int i=0; i<line_count; ++i) {
	 if(!line[i].enabled) continue;
	 if(usedPoints[i] > 1) {
	    pprintf("f  %d %d\n", i, usedPoints[i]);
	    gdk_draw_lines(dest, line[i].line_gc, &points[MAX_POINTS * i], usedPoints[i]);
	 }
      }

      pprintf("lc = %d\n", lc);

   }

   free(points);
   free(usedPoints);
   free(y);
   free(gotundef);

   while(gtk_events_pending()) gtk_main_iteration();

   plot_draw_legend(plot, dest, show_disabled, font_desc, w, h);
}

static void plot_draw_legend(my_plot_t *plot, GdkDrawable *dest, gboolean show_disabled, PangoFontDescription *font_desc, int w, int h) {

   // precalculate layouts to get max layout width
   int layout_max_size_x = 0;

   int list_count = 0;
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->line_set[lsi].dataset;
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      for(int i=0; i<line_count; ++i) {
	 if(!line[i].enabled && !show_disabled) continue;

	 const char *linename = dataset->item[line[i].dataset_idx].name;
	 char *text = malloc((dataset->name ? strlen(dataset->name) : 7 + 1 + 10) + 2 + (linename ? strlen(linename) : 4 + 1 + 10) + (line[i].mode == 1 ? 7 : 0) + 1);

	 if(plot->line_set_count > 1) {
	    if(dataset->name) {
	       strcpy(text, dataset->name);
	       strcat(text, ": ");
	    } else {
	       sprintf(text, "dataset %d: ", lsi + 1);
	    }
	 } else {
	    text[0] = '\0';
	 }

	 if(linename) {
	    strcat(text, linename);
	 } else {
	    sprintf(text + strlen(text), "item %d", i + 1);
	 }

	 switch(line[i].mode) {
	    case 0:
	       break;

	    case 1:
	       strcat(text, " (diff)");
	       break;
	 }

	 PangoLayout *layout = gtk_widget_create_pango_layout(plot->da, text);
	 pango_layout_set_font_description(layout, font_desc);

	 // overstrike if not enabled
	 if(!line[i].enabled) {
	    PangoAttribute *pa_st = pango_attr_strikethrough_new(1);
	    pa_st->start_index = 0;
	    pa_st->end_index = strlen(text);
	    PangoAttrList *pal = pango_attr_list_new();
	    pango_attr_list_insert(pal, pa_st);
	    pango_layout_set_attributes(layout, pal);
	 }

	 int lx, ly;
	 pango_layout_get_pixel_size(layout, &lx, &ly);

	 layout_max_size_x = max(layout_max_size_x, lx);

	 free(text);

	 line[i].legend_layout = layout;

	 list_count++;
      }
   }

   int legend_list_box_offset_x = 0;
   int legend_list_box_offset_y = 2;
   int legend_list_box_size_x = 8;
   int legend_list_box_size_y = 8;
   int legend_list_box_text_spacing_x = 5;

   int legend_list_unit_size_x = layout_max_size_x + legend_list_box_offset_x + legend_list_box_size_x + legend_list_box_text_spacing_x;
   int legend_list_unit_size_y = 13;

   int legend_list_size_x = legend_list_unit_size_x;
   int legend_list_size_y = legend_list_unit_size_y * list_count;

   int legend_margin_x = 10;
   int legend_margin_y = 10;

   int legend_list_padding_x = 8;
   int legend_list_padding_y = 5;

   int legend_corner_x = w - legend_margin_x - 2 * legend_list_padding_x - legend_list_size_x;
   int legend_corner_y = legend_margin_y;

   int legend_corner_x2 = legend_corner_x + 2 * legend_list_padding_x + legend_list_size_x;
   int legend_corner_y2 = legend_corner_y + 2 * legend_list_padding_y + legend_list_size_y;

   // draw box
   x_gdk_draw_rectangle(dest, plot->legend_bg_gc,     1, legend_corner_x+1, legend_corner_y+1, legend_corner_x2-1, legend_corner_y2-1);
   x_gdk_draw_rectangle(dest, plot->legend_border_gc, 0, legend_corner_x  , legend_corner_y  , legend_corner_x2  , legend_corner_y2  );

   plot->legend_x = legend_corner_x;
   plot->legend_y = legend_corner_y;
   plot->legend_x2 = legend_corner_x2;
   plot->legend_y2 = legend_corner_y2;

   int no = 0;

   // draw legend items
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      for(int i=0; i<line_count; ++i) {
	 if(!line[i].enabled && !show_disabled) continue;

	 int legend_list_corner_x = legend_corner_x + legend_list_padding_x;
	 int legend_list_corner_y = legend_corner_y + legend_list_padding_y + legend_list_unit_size_y * no;

	 line[i].legend_x = legend_list_corner_x;
	 line[i].legend_y = legend_list_corner_y;
	 line[i].legend_x2 = legend_list_corner_x + legend_list_unit_size_x;
	 line[i].legend_y2 = legend_list_corner_y + legend_list_unit_size_y;

	 int legend_list_text_corner_x = legend_list_corner_x + legend_list_box_offset_x + legend_list_box_size_x + legend_list_box_text_spacing_x;
	 int legend_list_text_corner_y = legend_list_corner_y;

	 x_gdk_draw_rectangle(dest, line[i].enabled ? line[i].line_gc : plot->legend_border_gc, line[i].enabled,
			      legend_list_corner_x + legend_list_box_offset_x,
			      legend_list_corner_y + legend_list_box_offset_y,
			      legend_list_corner_x + legend_list_box_offset_x + legend_list_box_size_x - 1,
			      legend_list_corner_y + legend_list_box_offset_y + legend_list_box_size_y - 1);

	 /*
	 x_gdk_draw_rectangle(dest, plot->legend_border_gc, 0,
			      legend_list_corner_x, legend_list_corner_y,
			      legend_list_corner_x + legend_list_unit_size_x, legend_list_corner_y + legend_list_unit_size_y);
	 */

	 gdk_draw_layout(dest, plot->legend_text_gc,
			 legend_list_text_corner_x,
			 legend_list_text_corner_y, line[i].legend_layout);

/*
	 if(!line[i].enabled) {
	    printf("%d %d\n", lx, ly);
	    gdk_draw_line(dest, plot->legend_text_gc,
			  legend_list_text_corner_x, legend_list_text_corner_y + ly/2,
			  legend_list_text_corner_x + lx, legend_list_text_corner_y + ly/2);
	 }
*/

	 no++;
      }
   }

   //    last outer2 if(normal_expose);

   //	    if(mytrans != plot->dsc_x->{num_changes} + plot->dsc_y->{num_changes}) {
   //	    print "- mytrans plot->dsc_x->{num_changes}\n";
   //	    da->queue_draw() unless(queued++);
   //	    queued++;
   //	    plot->conc--; return;

   //####################

   while(gtk_events_pending()) gtk_main_iteration();
}
