/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * util - misc functions
 */

#include <time.h>
#include <stdio.h>
#include <string.h>

#include "dataset.xh"
#include "stringutil.xh"
#include "util.xh"

const char *format_time(my_time_t time, int include_millis) {
   static char buff[4+1+2+1+2 +1+ 2+1+2+1+2 +1+ 3    +1+ 14  +1];
   time_t t = (time_t)(time / TIME_MULTIPLIER);
   struct tm *lt = localtime(&t);
   if(!strftime(buff, sizeof(buff), "%F %T", lt)) return "format_time error";
   if(include_millis) {
      char *p = buff + strlen(buff);
      snprintf(p, (size_t)(buff + sizeof(buff) - p), ".%03d", (int)((time % TIME_MULTIPLIER) * 1000 / TIME_MULTIPLIER));
   }
// snprintf(p, (size_t)(buff + sizeof(buff) - p), ".%03d %10d.%03d", (int)((time % TIME_MULTIPLIER) * 1000 / TIME_MULTIPLIER), (int)(time / TIME_MULTIPLIER), (int)((time % TIME_MULTIPLIER) * 1000 / TIME_MULTIPLIER));
   return buff;
}

const char *format_difftime(my_time_t time, int include_millis) {
   static char buff[5 +1+ 2+1+2+1+2 +1+ 3 +1];
   time_t tm = (time_t)(time / TIME_MULTIPLIER);
   int sec = tm % 60; tm /= 60;
   int min = tm % 60; tm /= 60;
   int hour = tm % 24; tm /= 24;
   snprintf(buff, sizeof(buff), "%d days %02d:%02d:%02d",
	    (int)tm, hour, min, sec);
   if(include_millis) {
      char *p = buff + strlen(buff);
      snprintf(p, (size_t)(buff + sizeof(buff) - p), ".%03d", (int)((tm % TIME_MULTIPLIER) * 1000 / TIME_MULTIPLIER));
   }
   return buff;
}

char *fgets_strip_comments(char *s, int size, FILE *stream) {
  char *line = fgets(s, size, stream);
  if(!line) return line;
  return strip_comments(line);
}

void x_gdk_draw_rectangle(GdkDrawable *drawable, GdkGC *gc, gboolean filled, gint x1, gint y1, gint x2, gint y2) {
   if(x1 > x2) {
      int t = x1;
      x1 = x2;
      x2 = t;
   }

   if(y1 > y2) {
      int t = y1;
      y1 = y2;
      y2 = t;
   }

   if(filled) {
      ++x2;
      ++y2;
   }

   gdk_draw_rectangle(drawable, gc, filled, x1, y1, x2-x1, y2-y1);
}

//void x_gdk_draw_arrow(GdkDrawable *drawable, GdkGC *gc, int fx, int fy, int tx, int ty, int arrow_side_len

int get_color(GdkColor *dst, const char *str) {
   // printf("get_color %s\n", str);
   GdkColormap *colormap = gdk_colormap_get_system();
   if(!gdk_color_parse(str, dst)) return 1;
/*
  // inverse
   dst->red ^= -1;
   dst->green ^= -1;
   dst->blue ^= -1;
*/
   return gdk_colormap_alloc_color(colormap, dst, 0, 0) ? 0 : 2;
}

GdkGC *gdk_gc_new_with_fg_color(GdkDrawable *drawable, const char *color) {
   GdkGC *gc = gdk_gc_new(drawable);
   // gdk_gc_copy(line[i].line_gc, plot->da->style->fg_gc[GTK_STATE_NORMAL]);
   GdkColor c;
   get_color(&c, color);
   gdk_gc_set_foreground(gc, &c);
   return gc;
}
